#include "Python.h"
#include "pyobjc-api.h"
#include <stdarg.h>

#import <Foundation/Foundation.h>

static CFStringRef aString = CFSTR("a static string");

@interface OC_PointerSupport : NSObject {
}
@end

@implementation OC_PointerSupport
+ (Py_ssize_t)getObjectLen:(PyObject*)value
{
    PyGILState_STATE state  = PyGILState_Ensure();
    Py_ssize_t       result = PyObject_Length(value);
    PyGILState_Release(state);
    return result;
}

+ (PyObject*)getNone
{
    PyGILState_STATE state = PyGILState_Ensure();
    Py_INCREF(Py_None);
    PyGILState_Release(state);
    return Py_None;
}

+ (Class)getClass
{
    return (Class)[OC_PointerSupport class];
}

+ (id)className:(Class)class
{
    if (class == Nil) {
        return @"No class given";
    }
    return [NSString
        stringWithUTF8String:(const char* _Nonnull)class_getName((Class) class)];
}

+ (CFStringRef)getString
{
    return aString;
}

+ (CFStringRef)getContext
{
    return (CFStringRef)kCFAllocatorUseContext;
}

@end

static PyObject*
make_opaque_capsule(PyObject* mod __attribute__((__unused__)))
{
    return PyCapsule_New((void*)1234, "objc.__opaque__", NULL);
}

static PyObject*
make_object_capsule(PyObject* mod __attribute__((__unused__)))
{
    NSObject* object = [[[NSObject alloc] init] autorelease];
    return PyCapsule_New(object, "objc.__object__", NULL);
}

static PyMethodDef mod_methods[] = {{
                                        "opaque_capsule",
                                        (PyCFunction)make_opaque_capsule,
                                        METH_NOARGS,
                                        0,
                                    },
                                    {
                                        "object_capsule",
                                        (PyCFunction)make_object_capsule,
                                        METH_NOARGS,
                                        0,
                                    },
                                    {0, 0, 0, 0}};

static int mod_exec_module(PyObject* m)
{
    if (PyObjC_ImportAPI(m) < 0) {
        return -1;
    }

    if (PyModule_AddObject(m, "OC_PointerSupport",
                           PyObjC_IdToPython([OC_PointerSupport class]))
        < 0) {
        return -1;
    }
    return 0;
}

static struct PyModuleDef_Slot mod_slots[] = {
    {
        .slot = Py_mod_exec,
        .value = (void*)mod_exec_module
    },
#if PY_VERSION_HEX >= 0x030c0000
    {
        /* This extension does not use the CPython API other than initializing
         * the module, hence is safe with subinterpreters and per-interpreter
         * GILs
         */
        .slot = Py_mod_multiple_interpreters,
        .value = Py_MOD_PER_INTERPRETER_GIL_SUPPORTED,
    },
#endif
#if PY_VERSION_HEX >= 0x030d0000
    {
        .slot = Py_mod_gil,
        .value = Py_MOD_GIL_NOT_USED,
    },
#endif
    {  /* Sentinel */
        .slot = 0,
        .value = 0
    }
};

static struct PyModuleDef mod_module = {
    .m_base = PyModuleDef_HEAD_INIT,
    .m_name = "pointersupport",
    .m_doc = NULL,
    .m_size = 0,
    .m_methods = mod_methods,
    .m_slots = mod_slots,
    .m_traverse = NULL,
    .m_clear = NULL,
    .m_free = NULL,
};

PyObject* PyInit_pointersupport(void);

PyObject* __attribute__((__visibility__("default"))) _Nullable PyInit_pointersupport(void)
{
    return PyModuleDef_Init(&mod_module);
}
